import { readFile } from 'node:fs/promises'
import { resolve } from 'node:path'
import { Rules, Rule, Predicate, Replace, Tape, TuringMachine, Seek } from './automaton.js'

const { First, Last, Left, Right, Stay, SeekFromStack, JumpFromStack } = Seek
const { ClearStack, Nothing, Push, Seq, Swap, With } = Replace
const { InBounds, Not, OutOfBounds, True } = Predicate

const State = Object.freeze({
    Start: 'start',
    ScanRight: 'scan-right',
    FindFile: 'find-file',
    ScanLeft: 'scan-left',
    FindSpace: 'find-space',
    ClearFile: 'clear-file',
})

const IsFreeSpace = Predicate.Equals('.')
const IsNotFree = Not(IsFreeSpace)
const IsFartherRight = x => x.peek().position <= x.position

const dot = ch => ch.symbol === '.'
const notDot = ch => !dot(ch)
const TrySkipFiles = x => {
    const stack = x.stack.slice()
    const a = stack.skipUntil(dot).takeUntil(notDot)
    const skip = !stack.done()
    //console.log(a, 'skipping?', skip, !stack.done())
    return skip
}

const FirstOrStack = x => {
    // unwind the stack
    const states = []
    let next
    while (next = x.pop()) {
        if (next.position < x.position) {
            break
        } else {
            states.push(next)
            next = null
        }
    }


    // rewind it
    states.reverse()
    let tmp
    while (tmp = states.pop()) {
        x.push(tmp)
    }

    return next ? next : First(x)
    //return x.peek() ? x.pop().position : First(x)
}

const TrySkipping = Seq(InBounds(Right), TrySkipFiles)

const HasMoreFiles = x => {
    const stack = x.stack.slice()
    stack.skipUntil(dot).skipUntil(notDot)
    return !stack.done()
}

const ShouldContinue = Predicate.Every(OutOfBounds(Right), HasMoreFiles)

//        |  From |  To |         Condition             | Next Character | Move Direction |
//        | ------------------------------------------------------------------------------|
// new Rule('q0',  'q1', reader => reader.symbol === '1', (_reader) => '0',     1) 

const rules = new Rules([
    new Rule(State.Start, State.ScanRight, True, Nothing, Right),
    new Rule(State.ScanRight, State.FindFile, TrySkipping, ClearStack, Last),
    new Rule(State.ScanRight, State.FindFile, ShouldContinue, ClearStack, Stay),
    new Rule(State.ScanRight, State.ScanRight, InBounds(Right), Push, Right),
    new Rule(State.FindFile, State.ScanLeft, IsNotFree, Seq(ClearStack, Push), Stay),
    new Rule(State.FindFile, State.FindFile, IsFreeSpace, Nothing, Left),
    new Rule(State.ScanLeft, State.ScanLeft, InBounds(Left), Nothing, FirstOrStack),
    new Rule(State.ScanLeft, State.FindSpace, OutOfBounds(Left), Nothing, Stay),
    new Rule(State.FindSpace, State.ScanRight, IsFartherRight, ClearStack, Right),
    new Rule(State.FindSpace, State.ClearFile, IsFreeSpace, Swap, JumpFromStack),
    new Rule(State.FindSpace, State.FindSpace, InBounds(Right), Nothing, Right),
    new Rule(State.ClearFile, State.ScanRight, True, With('.'), SeekFromStack(false)),
])

const char = x => String.fromCodePoint(x)
const toInt = x => x.codePointAt() - 32
const repeat = (n, value) => value.repeat(n)

const expand = input => {
    return input.split('').reduce((acc, n, i) => {
        if (i % 2) {
            return acc.concat(repeat(n, char(46)))
        } else {
            return acc.concat(repeat(n, char((i / 2) + 32)))
        }
    }, '')
}

const compress = input => {
    const tape = new Tape(input)
    const tm = new TuringMachine(tape, rules)
    tm.debug = false
    tm.exec()
    return tm.value
}

const notFree = ch => ch !== '.'
const mul = (a, b) => a * b
const add = (a, b) => a + b

const checksum = input => input.filter(notFree).map(toInt).map(mul).reduce(add)

const main = input => {
    const expanded = expand(input)
    console.log(expanded)
    const compressed = compress(expanded)
    console.log(compressed.filter(notFree).map(toInt))
    const sum = checksum(compressed)
    return sum
}

const tests = [
    ['2333133121414131402', 1928],
    ['1010101010101010101010', 385],
    ['111111111111111111111', 290],
    ['10101010101010101010101', 506],
    ['90909', 486]
]

tests.forEach(([input, expected]) => {
    console.log(`input length is ${input.length}`)
    const result = main(input)
    console.log(result === expected, result)
})

//readFile(resolve('./input'), { encoding: 'utf8' })
//    .then(x => {
//        console.log(`input length is ${x.length}`)
//        return main(x)
//    })
//    .then(console.log)

